use super::Token;

fn check_number(my_string: String, i: usize) -> Result<Token, String> {
    if my_string.as_str() == "." {
        return Err(format!("unexpected token: `.` at position {i}"));
    }
    let find = my_string.find(".");
    if let Some(j) = find {
        let find = my_string[j+1..].find(".");
        if let Some(_) = find {
            return Err(format!("unexpected token: `{my_string}` at position {i}"))
        }
    }
    Ok(Token::Number(my_string))
}

pub fn tokenize(query: &str) -> Result<Vec<Token>, String> {
    let mut tokens: Vec<Token> = vec![];
    let mut my_string = String::new();
    let mut is_last_number = false;
    let mut is_last_variable = false;

    for (i, token) in query.chars().enumerate() {
        let mut is_still_number = false;
        let mut is_still_variable = false;
        let mut is_pop_needed = true;
        match token {
            '+' => tokens.push(Token::Addition()),
            '-' => tokens.push(Token::Substraction()),
            '*' => tokens.push(Token::Multiplication()),
            '/' => tokens.push(Token::Division()),
            '%' => tokens.push(Token::Modulo()),
            '^' => tokens.push(Token::Exponentiation()),
            '=' => tokens.push(Token::Equal()),
            '(' => tokens.push(Token::OpenParenthesis()),
            ')' => tokens.push(Token::CloseParenthesis()),
            'a'..='z' | 'A'..='Z' => {
                if is_last_number == true {
                    tokens.push(check_number(my_string, i)?);
                    is_last_number = false;
                    my_string = String::new();
                }
                if is_last_variable == false {
                    is_last_variable = true;
                }
                is_still_variable = true;
                my_string += &token.to_string();
            },
            '0'..='9' | '.' => {
                if is_last_variable == true {
                    match my_string.as_str() {
                        "i" => tokens.push(Token::ImaginaryUnit()),
                        _ => tokens.push(Token::Variable(my_string)),
                    }
                    is_last_variable = false;
                    my_string = String::new();
                }
                if is_last_number == false {
                    is_last_number = true;
                }
                is_still_number = true;
                my_string += &token.to_string();
            },
            ' ' => is_pop_needed = false,
            _ => return Err(format!("unexpected token: `{token}` at position {i}")),
        };
        if is_last_variable && !is_still_variable {
            let mut tmp_token = None;
            if is_pop_needed {
                tmp_token = tokens.pop();
            }
            match my_string.as_str() {
                "i" => tokens.push(Token::ImaginaryUnit()),
                _ => tokens.push(Token::Variable(my_string)),
            }
            if let Some(tok) = tmp_token {
                tokens.push(tok);
            }
            is_last_variable = false;
            my_string = String::new();
        }
        else if is_last_number && !is_still_number {
            let mut tmp_token = None;
            if is_pop_needed {
                tmp_token = tokens.pop();
            }
            tokens.push(check_number(my_string, i)?);
            if let Some(tok) = tmp_token {
                tokens.push(tok);
            }
            is_last_number = false;
            my_string = String::new();
        }
    }
    if is_last_variable {
        match my_string.as_str() {
            "i" => tokens.push(Token::ImaginaryUnit()),
            _ => tokens.push(Token::Variable(my_string)),
        }
    }
    else if is_last_number {
        tokens.push(check_number(my_string, query.len())?);
    }

    Ok(tokens)
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn addition() {
        let query = "+";
        let result: Vec<Token> = vec![Token::Addition()];

        assert_eq!(tokenize(query).unwrap(), result);
    }

    #[test]
    fn substraction() {
        let query = "-";
        let result: Vec<Token> = vec![Token::Substraction()];

        assert_eq!(tokenize(query).unwrap(), result);
    }

    #[test]
    fn multiplication() {
        let query = "*";
        let result: Vec<Token> = vec![Token::Multiplication()];

        assert_eq!(tokenize(query).unwrap(), result);
    }

    #[test]
    fn division() {
        let query = "/";
        let result: Vec<Token> = vec![Token::Division()];

        assert_eq!(tokenize(query).unwrap(), result);
    }

    #[test]
    fn modulo() {
        let query = "%";
        let result: Vec<Token> = vec![Token::Modulo()];

        assert_eq!(tokenize(query).unwrap(), result);
    }

    #[test]
    fn exponentiation() {
        let query = "^";
        let result: Vec<Token> = vec![Token::Exponentiation()];

        assert_eq!(tokenize(query).unwrap(), result);
    }

    #[test]
    fn equal() {
        let query = "=";
        let result: Vec<Token> = vec![Token::Equal()];

        assert_eq!(tokenize(query).unwrap(), result);
    }

    #[test]
    fn open_parenthesis() {
        let query = "(";
        let result: Vec<Token> = vec![Token::OpenParenthesis()];

        assert_eq!(tokenize(query).unwrap(), result);
    }

    #[test]
    fn close_parenthesis() {
        let query = ")";
        let result: Vec<Token> = vec![Token::CloseParenthesis()];

        assert_eq!(tokenize(query).unwrap(), result);
    }

    #[test]
    fn imaginary_unit() {
        let query = "i";
        let result: Vec<Token> = vec![Token::ImaginaryUnit()];

        assert_eq!(tokenize(query).unwrap(), result);
    }

    #[test]
    fn variable() {
        let query = "variable";
        let result: Vec<Token> = vec![Token::Variable(String::from("variable"))];

        assert_eq!(tokenize(query).unwrap(), result);
    }

    #[test]
    fn variable_double_i() {
        let query = "ii";
        let result: Vec<Token> = vec![Token::Variable(String::from("ii"))];

        assert_eq!(tokenize(query).unwrap(), result);
    }

    #[test]
    fn number_natural() {
        let query = "123456";
        let result: Vec<Token> = vec![Token::Number(String::from("123456"))];

        assert_eq!(tokenize(query).unwrap(), result);
    }

    #[test]
    fn number_rational() {
        let query = "123.456";
        let result: Vec<Token> = vec![Token::Number(String::from("123.456"))];

        assert_eq!(tokenize(query).unwrap(), result);
    }

    #[test]
    fn number_point_something() {
        let query = ".123456";
        let result: Vec<Token> = vec![Token::Number(String::from(".123456"))];

        assert_eq!(tokenize(query).unwrap(), result);
    }

    #[test]
    fn number_trailing_point() {
        let query = "123456.";
        let result: Vec<Token> = vec![Token::Number(String::from("123456."))];

        assert_eq!(tokenize(query).unwrap(), result);
    }

    #[test]
    fn number_variable() {
        let query = "23x";
        let result: Vec<Token> = vec![Token::Number(String::from("23")), Token::Variable(String::from("x"))];
        assert_eq!(tokenize(query).unwrap(), result);

        let query = "23i";
        let result: Vec<Token> = vec![Token::Number(String::from("23")), Token::ImaginaryUnit()];
        assert_eq!(tokenize(query).unwrap(), result);

        let query = "x23";
        let result: Vec<Token> = vec![Token::Variable(String::from("x")), Token::Number(String::from("23"))];
        assert_eq!(tokenize(query).unwrap(), result);

        let query = "i23";
        let result: Vec<Token> = vec![Token::ImaginaryUnit(), Token::Number(String::from("23"))];
        assert_eq!(tokenize(query).unwrap(), result);
    }

    #[test]
    #[should_panic]
    fn number_double_point() {
        let query = "12.34.56";
        tokenize(query).unwrap();
    }

    #[test]
    #[should_panic]
    fn number_double_point_in_a_row() {
        let query = "123..456";
        tokenize(query).unwrap();
    }

    #[test]
    #[should_panic]
    fn number_only_point() {
        let query = ".";
        tokenize(query).unwrap();
    }

    #[test]
    #[should_panic]
    fn invalid_token() {
        let query = "324*43224+243_+234=234";
        tokenize(query).unwrap();
    }

    #[test]
    fn bug_pop_not_needed_between_variables_and_numbers() {
        let query = "5(7i)";
        let result: Vec<Token> = vec![Token::Number(String::from("5")), Token::OpenParenthesis(), Token::Number(String::from("7")), Token::ImaginaryUnit(), Token::CloseParenthesis()];
        assert_eq!(tokenize(query).unwrap(), result);
    }
}
